#include <Wire.h>
#include <MPU6050.h>
#include <Adafruit_NeoPixel.h>

// ----------------- USER SETTINGS -----------------

// Choose which LED effect to run on jump detection:
// 0 = Colorful Blink (flashes red, green, blue colors)
// 1 = Rainbow Wave (smooth rainbow animation across strip)
const int SELECTED_EFFECT = 0;

// Jump detection parameters:
// Lower threshold = more sensitive jump detection (may cause false triggers)
const float ACCEL_THRESHOLD = 400.0; // Acceleration change threshold to detect a jump
const unsigned long COOLDOWN_MS = 500; // Minimum delay between detections in milliseconds

// LED strip configuration
#define LED_PIN 6 // Pin connected to NeoPixel data input
#define LED_COUNT 30 // Number of LEDs in the strip

// Colorful blink effect settings
const int BLINK_TIMES = 3; // How many times to blink on jump
const unsigned long BLINK_ON_DURATION_MS = 50; // LED ON time per color in ms
const unsigned long BLINK_OFF_DURATION_MS = 25; // LED OFF time between colors in ms

// Rainbow wave effect settings
const int RAINBOW_CYCLES = 5; // Number of full rainbow cycles to animate
const unsigned long RAINBOW_STEP_DELAY_MS = 20; // Delay between animation frames in ms

// -------------------------------------------------

Adafruit_NeoPixel strip(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);
MPU6050 mpu;

// For jump detection timing and filtering
unsigned long lastTriggerTime = 0;
float filteredDiff = 0;
const float alpha = 0.1; // EMA smoothing factor for acceleration changes

// Baseline acceleration values (measured at startup)
int16_t baselineX, baselineY, baselineZ;
float baselineMagnitude = 0;

// Animation state tracking
bool isEffectRunning = false;
unsigned long effectTimer = 0;

// --- EFFECT MANAGEMENT STRUCTURE ---

// Function pointer types for animation control functions
typedef bool (*EffectStepFunc)(unsigned long now);

struct Effect {
const char* name; // Name shown in serial output
EffectStepFunc stepFunc; // Called repeatedly until effect is done
void (*startFunc)(); // Called once to initialize effect
};

// Forward declarations of effect functions
void startColorfulBlink();
bool stepColorfulBlink(unsigned long now);

void startRainbowWave();
bool stepRainbowWave(unsigned long now);

// Available effects
Effect EFFECTS[] = {
{"Colorful Blink", stepColorfulBlink, startColorfulBlink},
{"Rainbow Wave", stepRainbowWave, startRainbowWave},
};

const int NUM_EFFECTS = sizeof(EFFECTS) / sizeof(EFFECTS[0]);
int currentEffectIndex = -1;

// ------------------------------------
// Utility: Calculate vector magnitude
float magnitude(int16_t x, int16_t y, int16_t z) {
return sqrt((long)x * x + (long)y * y + (long)z * z);
}

// Measure baseline acceleration (device must be still during calibration)
void calibrateBaseline() {
long sumX = 0, sumY = 0, sumZ = 0;
const int samples = 50;

Serial.println("Calibrating baseline... Keep device still!");

for (int i = 0; i < samples; i++) {
int16_t ax, ay, az;
mpu.getAcceleration(&ax, &ay, &az);
sumX += ax;
sumY += ay;
sumZ += az;
delay(20);
}

baselineX = sumX / samples;
baselineY = sumY / samples;
baselineZ = sumZ / samples;
baselineMagnitude = magnitude(baselineX, baselineY, baselineZ);

Serial.print("Baseline magnitude: ");
Serial.println(baselineMagnitude);
}

void setup() {
Serial.begin(115200);
Wire.begin();

strip.begin();
strip.show(); // Initialize all pixels to off
strip.setBrightness(50);

mpu.initialize();
if (!mpu.testConnection()) {
Serial.println("MPU6050 connection failed! Check wiring.");
while (1);
}

Serial.println("MPU6050 ready!");
calibrateBaseline();

// Validate user effect selection
if (SELECTED_EFFECT < 0 || SELECTED_EFFECT >= NUM_EFFECTS) {
Serial.println("Invalid SELECTED_EFFECT, defaulting to 0");
currentEffectIndex = 0;
} else {
currentEffectIndex = SELECTED_EFFECT;
}
}

void loop() {
unsigned long now = millis();

// Read raw acceleration data
int16_t ax, ay, az;
mpu.getAcceleration(&ax, &ay, &az);

// Calculate magnitude and difference from baseline
long magSquared = (long)ax * ax + (long)ay * ay + (long)az * az;
if (magSquared < 0) magSquared = 0; // Safety check

float currentMag = sqrt(magSquared);
float diff = abs(currentMag - baselineMagnitude);

// Apply exponential moving average to smooth out noise
filteredDiff = alpha * diff + (1 - alpha) * filteredDiff;

// Reset filteredDiff if invalid (NaN or Inf)
if (isnan(filteredDiff) || isinf(filteredDiff)) {
filteredDiff = 0;
}

Serial.print("Filtered magnitude difference: ");
Serial.println(filteredDiff);

// Detect jump if threshold exceeded and cooldown elapsed
if (!isEffectRunning && filteredDiff > ACCEL_THRESHOLD && (now - lastTriggerTime > COOLDOWN_MS)) {
Serial.print("Jump detected! Starting effect: ");
Serial.println(EFFECTS[currentEffectIndex].name);
lastTriggerTime = now;
isEffectRunning = true;
EFFECTS[currentEffectIndex].startFunc();
}

// Run animation effect if active
if (isEffectRunning) {
bool done = EFFECTS[currentEffectIndex].stepFunc(now);
if (done) {
isEffectRunning = false;
strip.clear();
strip.show();
}
}

delay(20); // Loop delay for timing
}

// ------------------------
// Colorful Blink Effect
// ------------------------

uint32_t blinkColors[] = {
strip.Color(255, 0, 0), // Red
strip.Color(0, 255, 0), // Green
strip.Color(0, 0, 255) // Blue
};

int blinkStep = 0;

void startColorfulBlink() {
blinkStep = 0;
effectTimer = millis();
}

bool stepColorfulBlink(unsigned long now) {
int totalSteps = BLINK_TIMES * 6; // 3 colors * ON/OFF * times

if (blinkStep >= totalSteps) return true; // Animation finished

int stepIndex = blinkStep / 2; // Each color has ON and OFF step
bool onStep = (blinkStep % 2) == 0;

unsigned long duration = onStep ? BLINK_ON_DURATION_MS : BLINK_OFF_DURATION_MS;

if (now - effectTimer >= duration) {
effectTimer = now;

if (onStep) {
int colorIndex = stepIndex % 3;
strip.fill(blinkColors[colorIndex]);
strip.show();
} else {
strip.clear();
strip.show();
}

blinkStep++;
}

return false; // Animation ongoing
}

// ------------------------
// Rainbow Wave Effect
// ------------------------

int rainbowCycle = 0;
int rainbowHue = 0;

void startRainbowWave() {
rainbowCycle = 0;
rainbowHue = 0;
effectTimer = millis();
}

bool stepRainbowWave(unsigned long now) {
if (rainbowCycle >= RAINBOW_CYCLES) return true; // Animation done

if (now - effectTimer >= RAINBOW_STEP_DELAY_MS) {
effectTimer = now;

for (int i = 0; i < strip.numPixels(); i++) {
int pixelHue = rainbowHue + (i * 65536L / strip.numPixels());
uint32_t color = strip.gamma32(strip.ColorHSV(pixelHue));
strip.setPixelColor(i, color);
}
strip.show();

rainbowHue += 256;
if (rainbowHue >= 65536) {
rainbowHue = 0;
rainbowCycle++;
}
}

return false; // Animation ongoing
}